@@ -7,7 +7,7 @@
 // Absolutely no warranties, released under GNU GPL version 3 or at your 
 // option any later version.
 //
-// Copyright 2008 Enrico Tassi <tassi@cs.unibo.it>
+// Copyright 2008-2010 Enrico Tassi <gares@fettunta.org>
 
 #define _BSD_SOURCE
 #define _GNU_SOURCE
@@ -29,13 +29,20 @@
 # define O_NOATIME 0
 #endif
 
+#define STATIC static
+
 #define SHA_DIGEST_LENGTH 20
 
 #define __tostring(x) #x 
 #define tostring(x) __tostring(x)
 
-#define ERROR(cause, msg...) \
-	fprintf(stderr, "error [" tostring(cause) "]: " msg)
+#define ERROR(cause, msg...) { \
+	fprintf(stderr, "error [" tostring(cause) "]: " msg);\
+	fprintf(stdout, "ERROR " msg);\
+	exit(EXIT_FAILURE);\
+	}
+#define WARNING(cause, msg...) \
+	fprintf(stderr, "warning [" tostring(cause) "]: " msg)
 #define VERBOSE(cause,msg...) \
 	if (verbose) fprintf(stderr,"debug [" tostring(cause) "]: " msg)
 #define VERBOSE_NOH(msg...) \
@@ -46,9 +53,10 @@
 #define DEFAULT_MAIL_NUMBER 500000
 
 // int -> hex
-static char hexalphabet[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
+STATIC char hexalphabet[] = 
+	{'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
 
-int hex2int(char c){
+STATIC int hex2int(char c){
 	switch(c){
 		case '0': case '1': case '2': case '3': case '4':
 		case '5': case '6': case '7': case '8': case '9': return c - '0';
@@ -56,15 +64,15 @@ int hex2int(char c){
 		case 'd': case 'e': case 'f': return c - 'a' + 10;
 	}
 	ERROR(hex2int,"Invalid hex character: %c\n",c);
-	exit(1);
 }
 
-char tmpbuff_1[41];
-char tmpbuff_2[41];
-char tmpbuff_3[41];
-char tmpbuff_4[41];
+// temporary buffers used to store sha1 sums in ASCII hex
+STATIC char tmpbuff_1[SHA_DIGEST_LENGTH * 2 + 1];
+STATIC char tmpbuff_2[SHA_DIGEST_LENGTH * 2 + 1];
+STATIC char tmpbuff_3[SHA_DIGEST_LENGTH * 2 + 1];
+STATIC char tmpbuff_4[SHA_DIGEST_LENGTH * 2 + 1];
 
-char* txtsha(unsigned char *sha1, char* outbuff){
+STATIC char* txtsha(unsigned char *sha1, char* outbuff){
 	int fd;
 
 	for (fd = 0; fd < 20; fd++){
@@ -75,77 +83,125 @@ char* txtsha(unsigned char *sha1, char* outbuff){
 	return outbuff;
 }
 
-void shatxt(const char string[41], unsigned char outbuff[]) {
+STATIC void shatxt(const char string[41], unsigned char outbuff[]) {
 	int i;
 	for(i=0; i < SHA_DIGEST_LENGTH; i++){
 		outbuff[i] = hex2int(string[2*i]) * 16 + hex2int(string[2*i+1]);
 	}
 }
 
+// flags used to mark struct mail so that at the end of the scanning 
+// we output commands lookig that flag
 enum sight {
 	SEEN=0, NOT_SEEN=1, MOVED=2, CHANGED=3
 };
 
-static char* sightalphabet[]={"SEEN","NOT_SEEN","MOVED","CHANGED"};
+STATIC char* sightalphabet[]={"SEEN","NOT_SEEN","MOVED","CHANGED"};
 
-const char* strsight(enum sight s){
+STATIC const char* strsight(enum sight s){
 	return sightalphabet[s];
 }
 
+// since the mails and names buffers may be reallocated,
+// hashtables cannot record pointers to a struct mail or char.
+// they record the offset w.r.t. the base pointer of the buffers.
+// we define a type for them, so that the compiler complains loudly
+typedef size_t name_t;
+typedef size_t mail_t;
+
 // mail metadata structure
 struct mail {
 	unsigned char bsha[SHA_DIGEST_LENGTH]; 	// body hash value
 	unsigned char hsha[SHA_DIGEST_LENGTH]; 	// header hash value
-	char *name;    							// file name
-	enum sight seen;     				// already seen (means do not delete)
+	name_t __name;    						// file name, do not use directly
+	enum sight seen;     			        // already seen?
 };
 
 // memory pool for mail file names
-char *names;
-long unsigned int curname, max_curname, old_curname;
+STATIC char *names;
+STATIC name_t curname, max_curname, old_curname;
 
 // memory pool for mail metadata
-struct mail* mails;
-long unsigned int mailno, max_mailno;
+STATIC struct mail* mails;
+STATIC mail_t mailno, max_mailno;
 
 // hash tables for fast comparison of mails given their name/body-hash
-GHashTable *sha2mail;
-GHashTable *filename2mail;
-time_t lastcheck;
+STATIC GHashTable *bsha2mail;
+STATIC GHashTable *filename2mail;
+STATIC time_t lastcheck;
 
 // program options
-int verbose;
+STATIC int verbose;
+STATIC int dry_run;
+
+// ============================ helpers =====================================
+
+// mail da structure accessors
+STATIC struct mail* mail(mail_t mail_idx) {
+	return &mails[mail_idx];
+}
+
+STATIC char* mail_name(mail_t mail_idx) {
+	return &names[mails[mail_idx].__name];
+}
+
+STATIC void set_mail_name(mail_t mail_idx, name_t name) {
+	mails[mail_idx].__name = name;
+}
+
+// predicates for assert_all_are
+STATIC int directory(struct stat sb){ return S_ISDIR(sb.st_mode); }
+
+// stats and asserts pred on argv[optind] ... argv[argc-optind]
+STATIC void assert_all_are(
+	int(*predicate)(struct stat), char* description, char*argv[], int argc)
+{
+	struct stat sb;
+	int c, rc;
+	VERBOSE(input, "Asserting all input paths are: %s\n", description);
+	for(c = 0; c < argc; c++) { 
+		rc = stat(argv[c], &sb);
+		if (rc != 0) {
+			ERROR(stat,"unable to stat %s\n",argv[c]);
+		} else if ( ! predicate(sb) ) {
+			ERROR(stat,"%s in not a %s, arguments must be omogeneous\n",
+				argv[c],description);
+		}
+		VERBOSE(input, "%s is a %s\n", argv[c], description);
+	}
+}
+#define ASSERT_ALL_ARE(what,v,c) assert_all_are(what,tostring(what),v,c)
 
 // =========================== memory allocator ============================
 
-struct mail* alloc_mail(){
-	struct mail* m = &mails[mailno];
+STATIC mail_t alloc_mail(){
+	mail_t m = mailno;
 	mailno++;
 	if (mailno >= max_mailno) {
 		mails = realloc(mails, sizeof(struct mail) * max_mailno * 2);
 		if (mails == NULL){
 			ERROR(realloc,"allocation failed for %lu mails\n", max_mailno * 2);
-			exit(EXIT_FAILURE);
 		}
 		max_mailno *= 2;
 	}
 	return m;
 }
 
-void dealloc_mail(){
+STATIC void dealloc_mail(){
 	mailno--;
 }
 
 #define MAX_EMAIL_NAME_LEN 1024
 
-char *next_name(){
+STATIC char *next_name(){
 	return &names[curname];
 }
 
-char *alloc_name(){
-	char *name = &names[curname];
+STATIC name_t alloc_name(){
+	name_t name = curname;
+	size_t len = strlen(&names[name]);
 	old_curname = curname;
-	curname += strlen(name) + 1;
+	curname += len + 1;
 	if (curname + MAX_EMAIL_NAME_LEN > max_curname) {
 		names = realloc(names, max_curname * 2);
 		max_curname *= 2;
@@ -153,91 +209,102 @@ char *alloc_name(){
 	return name;
 }
 
-void dealloc_name(){
+STATIC void dealloc_name(){
 	curname = old_curname;
 }
 
 // =========================== global variables setup ======================
 
-guint sha_hash(gconstpointer key){
-	unsigned char * k = (unsigned char *) key;
+// convenience casts to be used with glib hashtables 
+#define MAIL(t) ((mail_t)(t))
+#define GPTR(t) ((gpointer)(t))
+
+STATIC guint bsha_hash(gconstpointer key){
+	mail_t m = MAIL(key);
+	unsigned char * k = (unsigned char *) mail(m)->bsha;
 	return k[0] + (k[1] << 8) + (k[2] << 16) + (k[3] << 24);
 }
 
-gboolean sha_equal(gconstpointer k1, gconstpointer k2){
-	if(!memcmp(k1,k2,SHA_DIGEST_LENGTH)) return TRUE;
+STATIC gboolean bsha_equal(gconstpointer k1, gconstpointer k2){
+	mail_t m1 = MAIL(k1);
+	mail_t m2 = MAIL(k2);
+	if(!memcmp(mail(m1)->bsha,mail(m2)->bsha,SHA_DIGEST_LENGTH)) return TRUE;
 	else return FALSE;
 }
 
+STATIC gboolean hsha_equal(gconstpointer k1, gconstpointer k2){
+	mail_t m1 = MAIL(k1);
+	mail_t m2 = MAIL(k2);
+	if(!memcmp(mail(m1)->hsha,mail(m2)->hsha,SHA_DIGEST_LENGTH)) return TRUE;
+	else return FALSE;
+}
+
+STATIC guint name_hash(gconstpointer key){
+	mail_t m = MAIL(key);
+	return g_str_hash(mail_name(m));
+}
+
+STATIC gboolean name_equal(gconstpointer k1, gconstpointer k2){
+	mail_t m1 = MAIL(k1);
+	mail_t m2 = MAIL(k2);
+	return g_str_equal(mail_name(m1), mail_name(m2));
+}
+
 // setup memory pools and hash tables
-void setup_globals(unsigned long int mno, unsigned int fnlen){
+STATIC void setup_globals(unsigned long int mno, unsigned int fnlen){
 	// allocate space for mail metadata
 	mails = malloc(sizeof(struct mail) * mno);
-	if (mails == NULL){
-		ERROR(malloc,"allocation failed for %lu mails\n",mno);
-		exit(EXIT_FAILURE);
-	}
-	mailno=0;
+	if (mails == NULL) ERROR(malloc,"allocation failed for %lu mails\n",mno);
+	
+	mailno=1; // 0 is reserved for NULL
 	max_mailno = mno;
 
 	// allocate space for mail filenames
 	names = malloc(mno * fnlen);
-	if (names == NULL){
+	if (names == NULL)
 		ERROR(malloc, "memory allocation failed for %lu mails with an "
 			"average filename length of %u\n",mailno,fnlen);
-		exit(EXIT_FAILURE);
-	}
+
 	curname=0;
 	max_curname=mno * fnlen;
 
 	// allocate hashtables for detection of already available mails
-	sha2mail = g_hash_table_new(sha_hash,sha_equal);
-	if (sha2mail == NULL) {
-		ERROR(sha2mail,"hashtable creation failure\n");
-		exit(EXIT_FAILURE);
-	}
+	bsha2mail = g_hash_table_new(bsha_hash,bsha_equal);
+	if (bsha2mail == NULL) ERROR(bsha2mail,"hashtable creation failure\n");
 
-	filename2mail = g_hash_table_new(g_str_hash,g_str_equal);
-	if (filename2mail == NULL) {
+	filename2mail = g_hash_table_new(name_hash,name_equal);
+	if (filename2mail == NULL) 
 		ERROR(filename2mail,"hashtable creation failure\n");
-		exit(EXIT_FAILURE);
-	}
 }
 
 // =========================== cache (de)serialization ======================
 
 // dump to file the mailbox status
-void save_db(const char* dbname, time_t timestamp){
-	long unsigned int i;
+STATIC void save_db(const char* dbname, time_t timestamp){
+	mail_t m;
 	FILE * fd;
 	char new_dbname[PATH_MAX];
 
 	snprintf(new_dbname,PATH_MAX,"%s.new",dbname);
 
 	fd = fopen(new_dbname,"w");
-	if (fd == NULL){
-		ERROR(fopen,"unable to save db file '%s'\n",new_dbname);
-		exit(1);
-	}
+	if (fd == NULL) ERROR(fopen,"unable to save db file '%s'\n",new_dbname);
 
-	for(i=0; i < mailno; i++){
-		struct mail* m = &mails[i];
-		if (m->seen == SEEN) {
+	for(m=1; m < mailno; m++){
+		if (mail(m)->seen == SEEN) {
 			fprintf(fd,"%s %s %s\n", 
-				txtsha(m->hsha,tmpbuff_1), txtsha(m->bsha,tmpbuff_2), 
-				m->name);
+				txtsha(mail(m)->hsha,tmpbuff_1), 
+				txtsha(mail(m)->bsha,tmpbuff_2), 
+				mail_name(m));
 		}
 	}
 
 	fclose(fd);
 
-	snprintf(new_dbname,PATH_MAX,"%s.mtime",dbname);
+	snprintf(new_dbname,PATH_MAX,"%s.mtime.new",dbname);
 
 	fd = fopen(new_dbname,"w");
-	if (fd == NULL){
-		ERROR(fopen,"unable to save db file '%s'\n",new_dbname);
-		exit(1);
-	}
+	if (fd == NULL) ERROR(fopen,"unable to save db file '%s'\n",new_dbname);
 
 	fprintf(fd,"%lu",timestamp);
 
@@ -245,7 +312,7 @@ void save_db(const char* dbname, time_t timestamp){
 }
 
 // load from disk a mailbox status and index mails with hashtables
-void load_db(const char* dbname){
+STATIC void load_db(const char* dbname){
 	FILE* fd;
 	int fields;
 	char new_dbname[PATH_MAX];
@@ -254,27 +321,26 @@ void load_db(const char* dbname){
 
 	fd = fopen(new_dbname,"r");
 	if (fd == NULL){
-		ERROR(fopen,"unable to load db file '%s'\n",new_dbname);
+		WARNING(fopen,"unable to load db file '%s'\n",new_dbname);
 		lastcheck = 0L;
 	} else {
 		fields = fscanf(fd,"%1$lu",&lastcheck);
-		if (fields != 1) {
+		if (fields != 1) 
 			ERROR(fscanf,"malformed db file '%s', please remove it\n",
-					new_dbname);
-			exit(EXIT_FAILURE);
-		}
+				new_dbname);
+
 		fclose(fd);
 	}
    
 	fd = fopen(dbname,"r");
 	if (fd == NULL) {
-		ERROR(fopen,"unable to open db file '%s'\n",dbname);
+		WARNING(fopen,"unable to open db file '%s'\n",dbname);
 		return;
 	}
 
 	for(;;) {
 		// allocate a mail entry
-		struct mail* m = alloc_mail();
+		mail_t m = alloc_mail();
 
 		// read one entry
 		fields = fscanf(fd,
@@ -288,23 +354,21 @@ void load_db(const char* dbname){
 		}
 		
 		// sanity checks
-		if (fields != 3) {
+		if (fields != 3)
 			ERROR(fscanf,"malformed db file '%s', please remove it\n",dbname);
-			exit(EXIT_FAILURE);
-		}
 
-		shatxt(tmpbuff_1, m->hsha);
-		shatxt(tmpbuff_2, m->bsha);
+		shatxt(tmpbuff_1, mail(m)->hsha);
+		shatxt(tmpbuff_2, mail(m)->bsha);
 
 		// allocate a name string
-		m->name = alloc_name();
+		set_mail_name(m,alloc_name());
 		
 		// not seen file, may be deleted
-		m->seen=NOT_SEEN;
+		mail(m)->seen=NOT_SEEN;
 
 		// store it in the hash tables
-		g_hash_table_insert(sha2mail,m->bsha,m);
-		g_hash_table_insert(filename2mail,m->name,m);
+		g_hash_table_insert(bsha2mail,GPTR(m),GPTR(m));
+		g_hash_table_insert(filename2mail,GPTR(m),GPTR(m));
 		
 	} 
 
@@ -314,79 +378,80 @@ void load_db(const char* dbname){
 // =============================== commands ================================
 
 #define COMMAND_SKIP(m) \
-	VERBOSE(skip,"%s\n",m->name)
+	VERBOSE(skip,"%s\n",mail_name(m))
 
 #define COMMAND_ADD(m) \
-	fprintf(stdout,"ADD %s %s %s\n",m->name, \
-		txtsha(m->hsha,tmpbuff_1), txtsha(m->bsha, tmpbuff_2))
+	fprintf(stdout,"ADD %s %s %s\n",mail_name(m), \
+		txtsha(mail(m)->hsha,tmpbuff_1), txtsha(mail(m)->bsha, tmpbuff_2))
 
 #define COMMAND_COPY(m,n) \
 	fprintf(stdout, "COPY %s %s %s TO %s\n",\
-		m->name,txtsha(m->hsha, tmpbuff_1),\
-		txtsha(m->bsha, tmpbuff_2),n->name)
+		mail_name(m),txtsha(mail(m)->hsha, tmpbuff_1),\
+		txtsha(mail(m)->bsha, tmpbuff_2),mail_name(n))
 
 #define COMMAND_COPYBODY(m,n) \
 	fprintf(stdout, "COPYBODY %s %s TO %s %s\n",\
-		m->name,txtsha(m->bsha, tmpbuff_1),\
-		n->name,txtsha(n->hsha, tmpbuff_2))
+		mail_name(m),txtsha(mail(m)->bsha, tmpbuff_1),\
+		mail_name(n),txtsha(mail(n)->hsha, tmpbuff_2))
 
 #define COMMAND_DELETE(m) \
-	fprintf(stdout,"DELETE %s %s %s\n",m->name, \
-		txtsha(m->hsha, tmpbuff_1), txtsha(m->bsha, tmpbuff_2))
+	fprintf(stdout,"DELETE %s %s %s\n",mail_name(m), \
+		txtsha(mail(m)->hsha, tmpbuff_1), txtsha(mail(m)->bsha, tmpbuff_2))
 	
 #define COMMAND_REPLACE(m,n) \
 	fprintf(stdout, "REPLACE %s %s %s WITH %s %s\n",\
-		m->name,txtsha(m->hsha,tmpbuff_1),txtsha(m->bsha,tmpbuff_2),\
-		txtsha(n->hsha,tmpbuff_3),txtsha(n->bsha,tmpbuff_4))
+		mail_name(m),txtsha(mail(m)->hsha,tmpbuff_1),\
+		txtsha(mail(m)->bsha,tmpbuff_2),\
+		txtsha(mail(n)->hsha,tmpbuff_3),txtsha(mail(n)->bsha,tmpbuff_4))
 
 #define COMMAND_REPLACE_HEADER(m,n) \
 	fprintf(stdout, "REPLACEHEADER %s %s %s WITH %s\n",\
-		m->name,txtsha(m->hsha,tmpbuff_1), txtsha(m->bsha,tmpbuff_2), \
-					txtsha(n->hsha,tmpbuff_3))
+		mail_name(m),txtsha(mail(m)->hsha,tmpbuff_1),\
+		txtsha(mail(m)->bsha,tmpbuff_2), \
+		txtsha(mail(n)->hsha,tmpbuff_3))
 
-// the hearth 
-void analize_file(const char* dir,const char* file) {    
+// the heart
+STATIC void analize_file(const char* dir,const char* file) {    
 	unsigned char *addr,*next;
 	int fd, header_found;
 	struct stat sb;
-	struct mail* alias, *bodyalias, *m;
+	mail_t alias, bodyalias, m;
 	GChecksum* ctx;
 	gsize ctx_len;
 
 	m = alloc_mail();
 	snprintf(next_name(), MAX_EMAIL_NAME_LEN,"%s/%s",dir,file);
-	m->name = alloc_name();
+	set_mail_name(m,alloc_name());
 
-	fd = open(m->name, O_RDONLY | O_NOATIME);
+	fd = open(mail_name(m), O_RDONLY | O_NOATIME);
 	if (fd == -1) {
-		ERROR(open,"unable to open file '%s'\n",m->name);
+		WARNING(open,"unable to open file '%s'\n",mail_name(m));
 		goto err_alloc_cleanup;
 	}
 
 	if (fstat(fd, &sb) == -1) {
-		ERROR(fstat,"unable to stat file '%s'\n",m->name);
+		WARNING(fstat,"unable to stat file '%s'\n",mail_name(m));
 		goto err_alloc_cleanup;
 	}
 
-	alias = (struct mail*)g_hash_table_lookup(filename2mail,m->name);
+	alias = MAIL(g_hash_table_lookup(filename2mail,GPTR(m)));
 
 	// check if the cache lists a file with the same name and the same
-	// mtime. if so, this is an old, untouched, message we can skip
-	if (alias != NULL && lastcheck > sb.st_mtime) {
-		alias->seen=SEEN;
+	// mtime. If so, this is an old, untouched, message we can skip
+	if (alias != 0 && lastcheck > sb.st_mtime) {
+		mail(alias)->seen=SEEN;
 		COMMAND_SKIP(alias);
 		goto err_alloc_fd_cleanup;
 	}
 
 	addr = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
 	if (addr == MAP_FAILED){
-		ERROR(mmap, "unable to load '%s'\n",m->name);
 		if (sb.st_size == 0) 
 			// empty file, we do not consider them emails
 			goto err_alloc_fd_cleanup;
 		else 
-			// mmap failed, XXX add more verbose output
-			exit(EXIT_FAILURE);
+			// mmap failed
+			ERROR(mmap, "unable to load '%s'\n",mail_name(m));
 	}
 
 	// skip header
@@ -399,7 +464,7 @@ void analize_file(const char* dir,const char* file) {
 	}
 
 	if (!header_found) {
-		ERROR(parse, "malformed file '%s', no header\n",m->name);
+		WARNING(parse, "malformed file '%s', no header\n",mail_name(m));
 		munmap(addr, sb.st_size);
 		goto err_alloc_fd_cleanup;
 	}
@@ -408,54 +473,54 @@ void analize_file(const char* dir,const char* file) {
 	ctx = g_checksum_new(G_CHECKSUM_SHA1);
 	ctx_len = SHA_DIGEST_LENGTH;
 	g_checksum_update(ctx, addr, next - addr);
-	g_checksum_get_digest(ctx, m->hsha, &ctx_len);
+	g_checksum_get_digest(ctx, mail(m)->hsha, &ctx_len);
 	g_checksum_free(ctx);
 
 	ctx = g_checksum_new(G_CHECKSUM_SHA1);
 	ctx_len = SHA_DIGEST_LENGTH;
 	g_checksum_update(ctx, next, sb.st_size - (next - addr));
-	g_checksum_get_digest(ctx, m->bsha, &ctx_len);
+	g_checksum_get_digest(ctx, mail(m)->bsha, &ctx_len);
 	g_checksum_free(ctx);
 
 	munmap(addr, sb.st_size);
 	close(fd);
 
-	if (alias != NULL) {
-		if(sha_equal(alias->bsha,m->bsha)) {
-			if (sha_equal(alias->hsha, m->hsha)) {
-				alias->seen = SEEN;
+	if (alias != 0) {
+		if(bsha_equal(GPTR(alias),GPTR(m))) {
+			if (hsha_equal(GPTR(alias), GPTR(m))) {
+				mail(alias)->seen = SEEN;
 				goto err_alloc_fd_cleanup;
 			} else {
 				COMMAND_REPLACE_HEADER(alias,m);
-				m->seen=SEEN;
-				alias->seen=CHANGED;
+				mail(m)->seen=SEEN;
+				mail(alias)->seen=CHANGED;
 				return;
 			}
 		} else {
 			COMMAND_REPLACE(alias,m);
-			m->seen=SEEN;
-			alias->seen=CHANGED;
+			mail(m)->seen=SEEN;
+			mail(alias)->seen=CHANGED;
 			return;
 		}
 	}
 
-	bodyalias = g_hash_table_lookup(sha2mail,m->bsha);
+	bodyalias = MAIL(g_hash_table_lookup(bsha2mail,GPTR(m)));
 
-	if (bodyalias != NULL) {
-		if (sha_equal(bodyalias->hsha,m->hsha)) {
+	if (bodyalias != 0) {
+		if (hsha_equal(GPTR(bodyalias), GPTR(m))) {
 			COMMAND_COPY(bodyalias,m);
-			m->seen=SEEN;
+			mail(m)->seen=SEEN;
 			return;
 		} else {
 			COMMAND_COPYBODY(bodyalias,m);
-			m->seen=SEEN;
+			mail(m)->seen=SEEN;
 			return;
 		}
 	}
 
 	// we should add that file
 	COMMAND_ADD(m);
-	m->seen=SEEN;
+	mail(m)->seen=SEEN;
 	return;
 
 	// error handlers, status cleanup
@@ -468,14 +533,11 @@ err_alloc_cleanup:
 }
 	
 // recursively analyze a directory and its sub-directories
-void analize_dir(const char* path){
+STATIC void analize_dir(const char* path){
 	DIR* dir = opendir(path);
 	struct dirent *dir_entry;
 
-	if (dir == NULL) {
-		fprintf(stderr, "Unable to open directory '%s'\n", path);
-		exit(EXIT_FAILURE);
-	}
+	if (dir == NULL) ERROR(opendir, "Unable to open directory '%s'\n", path);
 
 	while ( (dir_entry = readdir(dir)) != NULL) {
 		if (DT_REG == dir_entry->d_type){
@@ -505,7 +567,7 @@ void analize_dir(const char* path){
 	}
 }
 
-void analize_dirs(char* paths[], int no){
+STATIC void analize_dirs(char* paths[], int no){
 	int i;
 	for(i=0; i<no; i++){
 		// we remove a trailing '/' if any 
@@ -517,44 +579,35 @@ void analize_dirs(char* paths[], int no){
 
 // at the end of the analysis phase, look at the mails data structure to
 // identify mails that are not available anymore and should be removed
-void generate_deletions(){
-	long unsigned int i;
+STATIC void generate_deletions(){
+	size_t m;
 
-	for(i=0; i < mailno; i++){
-		struct mail* m = &mails[i];
-		if (m->seen == NOT_SEEN) 
+	for(m=1; m < mailno; m++){
+		if (mail(m)->seen == NOT_SEEN) 
 			COMMAND_DELETE(m);
 		else 
 			VERBOSE(seen,"STATUS OF %s %s %s IS %s\n",
-				m->name,txtsha(m->hsha,tmpbuff_1),
-				txtsha(m->bsha,tmpbuff_2),strsight(m->seen));
+				mail_name(m),txtsha(mail(m)->hsha,tmpbuff_1),
+				txtsha(mail(m)->bsha,tmpbuff_2),strsight(mail(m)->seen));
 	}
 }
 
-int extra_sha_file(const char* file) {    
+STATIC void extra_sha_file(const char* file, int suddenly_flush) {    
 	unsigned char *addr,*next;
 	int fd, header_found;
 	struct stat sb;
 	gchar* sha1;
 
 	fd = open(file, O_RDONLY | O_NOATIME);
-	if (fd == -1) {
-		ERROR(open,"unable to open file '%s'\n",file);
-		return 1;
-	}
+	if (fd == -1) ERROR(open,"unable to open file '%s'\n",file);
 
-	if (fstat(fd, &sb) == -1) {
-		ERROR(fstat,"unable to stat file '%s'\n",file);
-		close(fd);
-		return 2;
+	if (fstat(fd, &sb) == -1) ERROR(fstat,"unable to stat file '%s'\n",file);
+	if (! S_ISREG(sb.st_mode)) {
+		ERROR(fstat,"not a regular file '%s'\n",file);
 	}
 
 	addr = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
-	if (addr == MAP_FAILED){
-		ERROR(mmap, "unable to load '%s'\n",file);
-		close(fd);
-		return 3;
-	}
+	if (addr == MAP_FAILED) ERROR(mmap, "unable to load '%s'\n",file);
 
 	// skip header
 	for(next = addr, header_found=0; next + 1 < addr + sb.st_size; next++){
@@ -565,12 +618,7 @@ int extra_sha_file(const char* file) {
 		}
 	}
 
-	if (!header_found) {
-		ERROR(parse, "malformed file '%s', no header\n",file);
-		munmap(addr, sb.st_size);
-		close(fd);
-		return 4;
-	}
+	if (!header_found) ERROR(parse, "malformed file '%s', no header\n",file);
 
 	// calculate sha1
 	fprintf(stdout, "%s ", 
@@ -583,29 +631,8 @@ int extra_sha_file(const char* file) {
 	
 	munmap(addr, sb.st_size);
 	close(fd);
-	return 0;
-}
 
-
-int extra_sha_files(char* file[],int no) {    
-	int i;
-	struct stat sb;
-	for (i=0; i < no; i++){
-		const char * data = file[i];
-		int rc = stat(data, &sb);
-		if (rc != 0) {
-			ERROR(stat,"unable to stat %s\n",data);
-			exit(EXIT_FAILURE);
-		}
-		if ( S_ISREG(sb.st_mode) ){
-			rc = extra_sha_file(data);
-			if (rc != 0) return rc;
-		} else {
-			ERROR(stat,"not a regular file '%s'\n",data);
-			exit(EXIT_FAILURE);
-		}
-	}
-	return 0;
+	if (suddenly_flush) fflush(stdout);
 }
 
 // ============================ main =====================================
@@ -614,16 +641,17 @@ int extra_sha_files(char* file[],int no) {
 #define OPT_DB_FILE    301
 
 // command line options
-static struct option long_options[] = {
+STATIC struct option long_options[] = {
 	{"max-mailno", 1, NULL, OPT_MAX_MAILNO},
 	{"db-file"   , 1, NULL, OPT_DB_FILE}, 
 	{"verbose"   , 0, NULL, 'v'},
+	{"dry-run"   , 0, NULL, 'd'},
 	{"help"      , 0, NULL, 'h'},
 	{NULL        , 0, NULL, 0}, 
 };
 
 // command line options documentation
-const char* long_options_doc[] = {
+STATIC const char* long_options_doc[] = {
 	"Estimation of max mail message number (default " 
 		tostring(DEFAULT_MAIL_NUMBER) ")"
 		"\n                        " 
@@ -632,30 +660,30 @@ const char* long_options_doc[] = {
 		"automatically if needed", 
 	"Name of the cache for the endpoint (default db.txt)",
 	"Increase program verbosity (printed on stderr, short -v)", 
+	"Do not generate a new db file (short -d)",
 	"This help screen", 
 	NULL
 };
 
 // print help and bail out
-void help(char* argv0, int rc){
+STATIC void help(char* argv0){
 	int i;
 	char *bname = g_path_get_basename(argv0);
 
-	fprintf(stdout,"\nUsage: %s [options] paths...\n",bname);
+	fprintf(stdout,"\nUsage: %s [options] (dirs...|fifo)\n",bname);
 	for (i=0;long_options[i].name != NULL;i++) {
 		fprintf(stdout,"  --%-20s%s\n",
 			long_options[i].name,long_options_doc[i]);
 	}
 	fprintf(stdout,"\n\
-If paths is a list of regular files, %s outputs the sha1 of its header\n\
-and body separated by space.\n\n\
+If paths is a single fifo, %s reads from it file names and outputs the\n\
+sha1 of their header and body separated by space.\n\n\
 If paths is a list of directories, %s outputs a list of actions a client\n\
 has to perform to syncronize a copy of the same maildirs. This set of actions\n\
 is relative to a previous status of the maildir stored in the db file.\n\
 The input directories are traversed recursively, and every file encountered\n\
 inside directories named cur/ and new/ is a potential mail message (if it\n\
 contains no \\n\\n it is skipped).\n\n\
-Regular files and directories cannot be mixed in paths.\n\n\
 Every client must use a different db-file, and the db-file is strictly\n\
 related with the set of directories given as arguments, and should not\n\
 be used with a different directory set. Adding items to the directory\n\
@@ -664,7 +692,6 @@ are generated).\n", bname, bname);
 	fprintf(stdout,
 		"\nVersion %s, © 2009 Enrico Tassi, released under GPLv3, \
 no waranties\n\n",tostring(VERSION));
-	exit(rc);
 }
 
 int main(int argc, char *argv[]) {
@@ -678,9 +705,12 @@ int main(int argc, char *argv[]) {
 	time_t bigbang;
 
 	glib_check_version(2,16,0);
+	g_assert(MAIL(NULL) == 0);
+	g_assert(GPTR(0) == NULL);
+	g_assert(MAIL(GPTR(1)) == 1);
 
 	for(;;) {
-		c = getopt_long(argc, argv, "vh", long_options, &option_index);
+		c = getopt_long(argc, argv, "vhd", long_options, &option_index);
 		if (c == -1) break; // no more args
 		switch (c) {
 			case OPT_MAX_MAILNO:
@@ -692,38 +722,54 @@ int main(int argc, char *argv[]) {
 			case 'v':
 				verbose = 1;
 			break;
+			case 'd':
+				dry_run = 1;
+			break;
 			case 'h':
-				help(argv[0],EXIT_SUCCESS);
+				help(argv[0]);
+				exit(EXIT_SUCCESS);
 			break;
 			default:
-				help(argv[0],EXIT_FAILURE);
+				help(argv[0]);
+				exit(EXIT_FAILURE);
 			break;
 		}
 	}
 
-	if (optind >= argc) help(argv[0],EXIT_FAILURE);
+	if (optind >= argc) { 
+		help(argv[0]);
+		exit(EXIT_FAILURE);
+	}
 
 	// remaining args is the dirs containing the data or the files to hash
 	data = argv[optind];
 
 	// check if data is a directory or a regular file
 	c = stat(data, &sb);
-	if (c != 0) {
-		ERROR(stat,"unable to stat %s\n",data);
-		exit(EXIT_FAILURE);
-	}
-	if ( S_ISREG(sb.st_mode) ){
-		// simple mode, just hash the file
-		return extra_sha_files(&argv[optind],argc - optind);
+	if (c != 0) ERROR(stat,"unable to stat %s\n",data);
+	
+	if ( S_ISFIFO(sb.st_mode) && argc - optind == 1){
+		FILE *in = fopen(data,"r");
+		if (in == NULL) {
+			ERROR(fopen,"unable to open fifo %s\n",data);
+			exit(EXIT_FAILURE);
+		}
+		while (!feof(in)) {
+			char name[MAX_EMAIL_NAME_LEN];	
+			if(fgets(name,MAX_EMAIL_NAME_LEN,in) != NULL){
+				size_t len = strlen(name);
+				if (len > 0 && name[len-1] == '\n') name[len-1]='\0';
+				extra_sha_file(name,1);	
+			}
+		}
+		exit(EXIT_SUCCESS);
 	} else if ( ! S_ISDIR(sb.st_mode) ) {
-		ERROR(stat, "given path is not a regular file nor a directory: %s\n",
-			data);
-		exit(EXIT_FAILURE);
+		ERROR(stat, "given path is not a fifo nor a directory: %s\n",data);
 	}
 	
-	VERBOSE(init,"data directories are ");
-	for(c=optind; c<argc - optind; c++) { VERBOSE_NOH(argv[c]);}	
-	VERBOSE_NOH("\n");
+	// regular case, hash the content of maildirs rooted in the 
+	// list of directories specified at command line
+	ASSERT_ALL_ARE(directory, &argv[optind], argc - optind);
 
 	// allocate memory
 	setup_globals(mailno,filenamelen);
@@ -735,7 +781,7 @@ int main(int argc, char *argv[]) {
 
 	generate_deletions();
 
-	save_db(dbfile, bigbang);
+	if (!dry_run) save_db(dbfile, bigbang);
 
 	exit(EXIT_SUCCESS);
 }
